/* sd.c
   Provides an inferace to an SD card over an SPI port.
   
   Author: David Collins
   
   Reference:
   Robert Imhoff, info@flashgenie.net
   http://www.flashgenie.net
   
*/

#include <hidef.h> 
#include <mc9s12c32.h>

#include "sd.h"
#include "debug.h"

// global variable for return values
unsigned char ret;

// set by write command which didn't complete
unsigned char SDbusy = 0;

// internal function called to wait for write or erase operations to end
// returns 0 on success
unsigned char SD_BusyWait(unsigned long waitCycles);


// send out 2 empty characters to disengage SD card
#define SD_Off() ret = SPI_ShiftByte(SD_EMPTY_CHAR); ret = SPI_ShiftByte(SD_EMPTY_CHAR); SD_DESELECT();


// number of bytes to send while SD card is busy before giving up
#define SD_WRITE_WAIT_TIMEOUT 50000
#define SD_CMD_TIMEOUT 520
#define SD_CMD_RESP_TIMEOUT 10
#define SD_INIT_TIMEOUT 2000
#define SD_INIT_INSERT_TIMEOUT 10

#define SD_READ_BUSY_WAIT 32000
#define SD_READ_RETRIES 5
#define SD_WRITE_RETRIES 5
#define SD_WRITE_ACK_BUSY_WAIT 100



// Initialize microcontroller registers for use with SD card  ** Device Specific **
void SD_Init(void) 
{
   // PTM5 - SPI Clock
   // PTM4 - SPI master output
   // PTM3 - Chip select bar - controlled separately in code
   // PTM2 - SPI slave input
   
   // PTM1 - Card inserted when low
 
 
   // I don't think these are required, enabling SPI should force them to be as needed
   DDRM = 0b00111000;  // PTM5-3 = output  
   PTM = 0;          // zero out PTM registers
   SD_DESELECT();
   
   // set PRM1 to have a pull up
   PERM = 0b00000010;
   PPSM = 0b00000000;
   
   
   
   // enable SPI, set to master, use clock normally low, input on rising edge of clock
   // MSB first send
   // do not use chip select in the normal way because it needs to stay low throughout the entire operation
   //  enabling it here would make it go high after every byte
   // CPOL=0, CPHA=0
   SPICR1 = 0b01010000;
   
  
   // PTM3 is chip select, but not with CPU control, no bidirectional stuff or wait mode
   SPICR2 = 0b00000000;
  
  
   // set the clock rate
   /* SPIBR is 0 M2 M1 M0 0 E2 E1 E0
      divisor = (M2M1M0 + 1) * 2 ^ (E2E1E0 + 1)
      SD has to run at <=400kHz at the beginning of initialization, then up to 25MHz afterward
      with a bus clock of 24MHz, that cooresponds to SPIBR = 0b00000101
   */
   
   SD_SET_SPEED_LOW();
   SDbusy = 0;
   
   // clear SPIF
   ret = SPISR;
   ret = SPIDR;   
}




// proceeds though an entire SPI shift operation and returns the value
// sent from the slave
unsigned char SPI_ShiftByte(unsigned char input) 
{
   SPI_WAIT_FOR_SPACE();
   SPIDR = input;        // put input byte in the SPI output buffer
   SPI_WAIT_FOR_DATA();
   return SPIDR;         // return the value received
}



// get two-byte R2 response from SD card if write command fails
unsigned short SD_GetStatus(void)          
{
   unsigned char r;
   unsigned short status;
  
   // CMD13 - status command  
   r = SD_SendCommand(13, 0);  
   status = ((unsigned short)r) << 8; 
  
   // receive second byte of response           
   r = SPI_ShiftByte(SD_EMPTY_CHAR); 
   status |= (unsigned short)r;
              
   return status;
}




// internal function called to wait for write or erase operations to end
// returns 0 on success
unsigned char SD_BusyWait(unsigned long waitCycles)
{
   unsigned long i;
   unsigned char b;
 
   DEBUG("Waiting for write sequence to finish...\r\n");
   // send waitCycles * 8 clock pulse at max
   for (i = 0; i < waitCycles; i++)
   {   
      b = SPI_ShiftByte(SD_EMPTY_CHAR);    
      if (b) // non-zero value: card came out of busy state
      {				            
          SDbusy = 0;
          return 0;
      }       
   }
   
   //only get here with failure, still busy
   SD_Off();      
   return 1; 
}




// send an SD command code to the SD card
unsigned char SD_SendCommand(unsigned char command, unsigned long blocknum) 
{
   unsigned char crc, b;
   unsigned long i, address;
  
   // set the framing bits properly
   command |= 0x40;
   command &= ~0x80;
  
   // this crc code corresponds to CMD0, but SPI mode in the SD card
   // ignores the CRC after CMD0 completes, therefore, 0x95 will always be used
   crc = 0x95;
  
   SD_SELECT();  // assert chip select
  
  
   // Wake up the SD card by sending it 0xFF repeatedly
   if (SDbusy) //busy flag was set by previous write or erase command, need longer wait
   {
      // wait for any pending write/erase operation to terminate
      b = SD_BusyWait(SD_WRITE_WAIT_TIMEOUT);
      
      // error -- card did not come out of busy state      
      if (b) return b;    
   } 
   else //busy flag was not set, but still have to wake up card
   {      
      for (i = 0; i < SD_CMD_TIMEOUT; i++) // send 0xff till non-0 answer (0 = busy)
      {
         b = SPI_ShiftByte(SD_EMPTY_CHAR);   //  clock cycles to wake up if needed
         if (b == 0xff) break;             // no longer busy         
      }
   }
   
   // convert the block number into a 32bit address
   address = blocknum * SD_BLOCK_SIZE;

   // send the actual command
   b = SPI_ShiftByte(command);
   // address sent MSB first
   b = SPI_ShiftByte((unsigned char)(address >> 24));
   b = SPI_ShiftByte((unsigned char)(address >> 16));
   b = SPI_ShiftByte((unsigned char)(address >> 8));
   b = SPI_ShiftByte((unsigned char)address);
   b = SPI_ShiftByte(crc);
 
   // send 0xFF until non-FF answer
   for (i = 0; i < SD_CMD_RESP_TIMEOUT; i++)       
   {
      if (b != SD_EMPTY_CHAR)   // have response: return
          return b;             // return command response byte
      b = SPI_ShiftByte(SD_EMPTY_CHAR);
   }
   
   // CMD timeout, only arrive here if failure
   return (SD_EMPTY_CHAR);
}


// Send the initialization sequence to the SD card
// returns 0 if successful
unsigned char SD_Card_Init(void) 
{
   int i, j, k;
   unsigned char r;
  
   
   SD_DESELECT();
   
   // initialization must be done with SCLK <= 400kHz
   SD_SET_SPEED_LOW();
   
   DEBUG("SD init...\r\n");
  
   // attempt to initialize the card several times
   for (k=0; k<SD_INIT_INSERT_TIMEOUT; k++)
   {
      if (SD_IS_CARD_INSERTED()) // only proceed if the card is inserted
      {       
         for (j=0; j<SD_INIT_TIMEOUT; j++)
         {
            //DEBUG(" Sending dummy pulses\r\n");
            // send dummy pulses to get SD card into native mode	  
            // must receive at least 74 clock ticks to initialize internal registers            
            for (i = 0; i < 12; i++)
            {                 
               r = SPI_ShiftByte(SD_EMPTY_CHAR);
            }
            SD_SELECT();       // activate chip
            
            // send CMD0 to switch to SPI mode
            r = SD_SendCommand(0, 0);
            
            if (r <= 1)  //have answer
               break;
            // else wait some more
            SD_DESELECT();            
         }
         
         // try ACMD55 -- MMC will respond with error
         r = SD_SendCommand(55, 0);  
              
         if (r == 5)
         {            
            // *** MMC ***
            // CardType = MMCType;
            
            // loop until card finishes initializing
            for (i=0; i<SD_INIT_TIMEOUT; i++) 
            {
               // CMD1 :  returns 0x01 while the card is initializing, 0 when finished          
               r = SD_SendCommand(1, 0);
               if (r==0) break;      
            }
         }
         else
         {
            // *** SD ***
            // CardType = SDType;
            
            // ACMD41 complete the ACMD
            r = SD_SendCommand(41, 0);
            
            DEBUG(" Sending CMD55, ACMD41 until initialized\r\n");
            // loop until card finishes initializing
            for (i=0; i<SD_INIT_TIMEOUT; i++) 
            {
               // CMD55
               r = SD_SendCommand(55, 0); 
               // ACMD41 : returns 0x01 while the card is initializing, 0 when finished
               r = SD_SendCommand(41, 0);
               if (r==0) break;      
            }
            
            DEBUG(" SD card clock speed set high\r\n");
            SD_SET_SPEED_HIGH();
         }
			         
         if (r == 0)    // success
         {       
            DEBUG(" SD card successfully initialized\r\n");     
            SD_Off();
            return 0;
         }
         else // stay in loop to try again
         {
            // did not finish initializing
            DEBUG("Retrying SD init.\r\n");
            // Display_CRLF(); Display_Str("retry init ");
            
         }
      }
   }

   // only get here if everything timed out
   return r;
}


// Read in a block from the SD card
// Returns 0 on success
unsigned char SD_Read_Block(unsigned char *buffer, unsigned long blocknum)
{   
   int i;
   unsigned char r;
   unsigned long n;
   
   if (!SD_IS_CARD_INSERTED())
   {
      DEBUG("Could not read, no card inserted.\r\n");
      
      SD_DESELECT();
      return 1;
   }

   // send out read command until it is accepted
   for (i=0; i<SD_READ_RETRIES; i++)
   {	     
      // CMD17 - read block  
      r = SD_SendCommand(17, blocknum); 
      if (r == 0)
      {
         // succeeded
         break;
      }
      else
      {  
         // try to resynch clock
         r = SPI_ShiftByte(SD_EMPTY_CHAR);
         
         // these two lines might not be needed
         SD_DESELECT();         
         SD_SELECT();
         
         r = SPI_ShiftByte(SD_EMPTY_CHAR);
      }
   }
   
   if (r) //  error...
   {
      // SD card did not accept read command                          
      DEBUG("Failed to read from SD card.\r\n");
      
      SD_Off();      
      // fail
      return r;
   }
   
   // read block -- response ok */
   // send 0xff till non-empty answer with data start token */
   for (n=0; n<SD_READ_BUSY_WAIT; n++)
   {
      r = SPI_ShiftByte(SD_EMPTY_CHAR);
      if (r != SD_EMPTY_CHAR)
         break;
   }
   
   // 0xFE is the start data token
   // read failed if it is not received
   if (r != SD_DATA_START_TOKEN)
   {
      DEBUG("SD read data token not received.\r\n");
      // CMD12 - Stop read
      r = SD_SendCommand(12, 0);
      SD_Off();
      return 1;
   }

   // read ok: data start token received
   // start reading data block
   for (i=0; i<SD_BLOCK_SIZE; i++)
   {
      *buffer = SPI_ShiftByte(SD_EMPTY_CHAR);
      buffer++;
   }
   
   // the SD card sends 2 CRC bytes after the data block
   // receive them, but ignore them
   r = SPI_ShiftByte(SD_EMPTY_CHAR);
   r = SPI_ShiftByte(SD_EMPTY_CHAR);
   
   DEBUG("SD read succeeded.\r\n");
   SD_Off();
   return 0;
}

							  
							  
							  
// Read in a block from the SD card
// Returns 0 on success
unsigned char SD_Write_Block(unsigned char *buffer, unsigned long blocknum)
{   
   int i;
   unsigned char r;   
   unsigned short r2;


   if (!SD_IS_CARD_INSERTED())
   {
      DEBUG("Could not write, no card inserted.\r\n");
      
      SD_DESELECT();
      return 1;
   }

   // send out read command until it is accepted
   for (i=0; i<SD_WRITE_RETRIES; i++)
   {	     
      // CMD24 - write block  
      r = SD_SendCommand(24, blocknum); 
      if (r == 0)
      {
         // succeeded
         break;
      }
      else
      {  
         // try to resynch clock
         r = SPI_ShiftByte(SD_EMPTY_CHAR);
         
         // these two lines might not be needed
         SD_DESELECT();         
         SD_SELECT();
         
         r = SPI_ShiftByte(SD_EMPTY_CHAR);
      }
   }
   
   if (r) //  error...
   {
      // SD card did not accept write command                          
      DEBUG("Failed to write to SD card.\r\n");
      
      SD_Off();      
      // fail
      return r;
   }


   //  response ok -- write the block
   
   // send the data start token
   r = SPI_ShiftByte(SD_EMPTY_CHAR);
   r = SPI_ShiftByte(SD_DATA_START_TOKEN);

   // finally, send the actual data
   for (i=0; i<SD_BLOCK_SIZE; i++)
   {
      r = SPI_ShiftByte(*buffer);
      buffer++;
   }
   
   // the SD card expects 2 CRC bytes after the data block
   // send 2 filler bytes
   r = SPI_ShiftByte(SD_EMPTY_CHAR);
   r = SPI_ShiftByte(SD_EMPTY_CHAR);
   

   for (i=0; i<SD_WRITE_ACK_BUSY_WAIT; i++)
   {
      // Send dummy bytes until response byte received
      r = SPI_ShiftByte(SD_EMPTY_CHAR);
               
      if ((r != SD_EMPTY_CHAR) && (r != SD_BUSY_CHAR))
         break;
   }

   // time out because no response
   if (r == SD_EMPTY_CHAR || r == SD_BUSY_CHAR)
   {                                
      SD_Off();
      DEBUG("SD card write timed out.\r\n");
      return 1;
   }
   
   // else interpret data response
   r &= 0x1F;        
   if (r != 0x05)                  
   {  
      //  rejection, get next two bytes to return error code
      r2 = SD_GetStatus();
      SD_Off();
      DEBUG("Write error occured.\r\n");
      
      // change to be unsigned short return
      return (unsigned char)r2;
   }
   
   // else data accepted
   // device stays busy to write block
   SDbusy = 1;
   DEBUG("SD write succeeded.\r\n");
   
   return 0;
}
